/*
 *
 * Most Recent Version 29-oct-97:
 *   - No debugging output, no printmatrix
 *
 *
 * Normal C version of Eigenvalue program by Soumen etc al
 *
 * Stefan Haenssgen  02-jul-95
 *
 * 02-jul-95	Removed task queue & runtime stuff
 *		Reformatted source text
 * 27-nov-96	Reused for c thread experiments
 *		Added random number generation -> FillMatrixRnd
 *		1st argument now determines initialization of matrix
 *		Added instrumentation for profiling
 *
 */

#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <stdlib.h>             /* Solaris, for rand() */
#include <sys/types.h>
#include <sys/time.h>
#include <sys/timeb.h>

extern int rand();
extern void srand();


#define SAFEMN (1.0e-307)	/* numerical stuff */
#define ULP    (2.22e-16)
#define FUDGE  (2.0)
#define MAX(a,b) ((a)>(b)?(a):(b))
#define MIN(a,b) ((a)<(b)?(a):(b))

#define MAXRND ((2<<15)-1)	/* maximum rnd number from rand() */

typedef struct {
    int depth;			/* task */
    int lc, rc;
    double lv, rv;
} Interval_t;


/*
 * proc-local static vars
 */

int n;				/* side of matrix */
double * d, * e;		/* diagonal and off-diagonal */


/* 
 * FillMatrix*                     EIGEN stuff
 *---------------------------------------------------------------------------
 * Generate a input matrix with geometrically distributed eigenvalues
 * writes into d[] and e[]
 */

void FillMatrixGeom ( int n, double *d, double *e )
{
    int i;
    double fac;

    for ( i=0; i<n; i++)
	*d++ = 3.0;
    for ( i=0; i<n-1; i++)
	*e++ = 1.0;

    *e++ = 0.0, fac = 0.98;	/* [sth] added ++ !! */
    d -= n, e -= n;		/* BUG?? leaves e index at -1 ! */

    for ( i=0; i<n; i++ ) {
	*d *= fac, *e *= fac;
	d++, e++, fac *= 0.98;
    }
    d -= n, e -= n;
    fac = 0.98;

    for ( i=0; i<n-1; i++ ) {
	*d *= fac, *e *= fac;
	d++, e++, fac *= 0.98;
    }
    *d *= fac;
    
    d -= n-1, e -= n-1;		/* [sth] reset indices for printing */
    
}


void FillMatrixUnif ( int n, double *d, double *e )
{
    int i, j;

    for ( i=0; i<n; i++ )
            *d++ = (double)i + ((double)(i/64))*(double)64/(double)n;
    for ( i=0; i<n-1; i++ )
        *e++ = 1.0e-9;
    *e = 0.0;

    d -= n, e -= n-1;           /* [sth] reset indices for printing */

}  


void FillMatrixRnd ( int n, double *d, double *e )
{
    int i, j;

    for ( i=0; i<n; i++ )
	*d++ = (double) rand() / (double) MAXRND;
    for ( i=0; i<n-1; i++ )
	*e++ = (double) rand() / (double) MAXRND;
    *e++ = 0.0;

    d -= n-1, e -= n-1;		/* [sth] reset indices for printing */
    
}



/*
 * Gerschgorin(n, d, e, &gl, &gu)
 *----------------------------------------------------------------------------
 * compute initial limits
 */
void Gerschgorin(int n, double *dIn, double *eIn, double * pgl, double *pgu)
{
    double pivMin, tmp, tmp1, tmp2, bnorm;
    double *d, *e;
    int i;

    for ( i=0, d=dIn, e=eIn, pivMin=1.0; i<n-1; i++, d++, e++ ) {
	tmp = (*e) * (*e);
	if ( fabs((*d)*(*(d+1))*ULP*ULP + SAFEMN) < tmp )
	    pivMin = MAX(pivMin,tmp);
    }
    pivMin *= SAFEMN;

    tmp1 = 0.0,  *pgl = *pgu = *dIn;
    for ( i=0, d=dIn, e=eIn; i<n-1; i++, d++, e++ ) {
	tmp2 = fabs(*e);
	*pgu = MAX(*pgu,*d+tmp1+tmp2);
	*pgl = MIN(*pgl,*d-tmp1-tmp2);
	tmp1 = tmp2;
    }
    *pgu = MAX(*pgu,*d+tmp1);
    *pgl = MIN(*pgl,*d-tmp1);
    bnorm = MAX(fabs(*pgl),fabs(*pgu));
    *pgl = *pgl - FUDGE*bnorm*ULP*n - FUDGE*pivMin;
    *pgu = *pgu + FUDGE*bnorm*ULP*n + FUDGE*pivMin;
}



/*
 * IeeeCount
 *---------------------------------------------------------------------------
 */

int IeeeCount ( int n, double *d, double *e, double sigma )
{
    double tmp, *pd, *pe2;
    int count, i;

/*    printf("IeeeCount(%d,d,e,%e) = ",n,sigma);*/
    
    pd = d, pe2 = e;
    tmp = *pd++ - sigma;
    count = ((*((int *)&tmp)) >> 31) & 1;
    for ( i=1; i<n; i++ ) {
	tmp = *pd++ - *pe2++/tmp - sigma;
	count += ((*((int *)&tmp)) >> 31) & 1;
    }

/*    printf("%d\n",count);*/

    return count;
}


/*
 * IsLeaf
 *--------------------------------------------------------------------------
 */

int IsLeaf ( double left, double right, double abstol, double reltol )
{
    double diff, mag;

/*    printf("IsLeaf(%e,%e,%e,%e) =",left,right,abstol,reltol);*/
    
    diff = fabs(left - right);
    mag = MAX(fabs(left), fabs(right));
    
/*    printf("%s\n",(diff < MAX(abstol, reltol * mag))?"yes":"no");*/
    
    return diff < MAX(abstol, reltol * mag);
}



void LoopBody_h ( Interval_t * new )
{
    int mc;
    double mv;

				/* new contains valid interval */

    do {
	mv = (new->lv + new->rv)/2.0;
	mc = IeeeCount(n, d, e, mv);
	mc = MIN(mc, new->rc);
	mc = MAX(mc, new->lc);
	if ( mc == new->lc )
	    new->lv = mv;
	if ( mc == new->rc )
	    new->rv = mv;
    } while ( ! IsLeaf(new->lv, new->rv, ULP, 2.0*ULP)
	      && ( mc == new->lc || mc == new->rc ) );

    if ( IsLeaf(new->lv, new->rv, ULP, 2.0*ULP) ) {
    }
    else {			/* have two children */
	Interval_t * lkid = malloc(sizeof(Interval_t));
	Interval_t * rkid = malloc(sizeof(Interval_t));
	rkid->depth = new->depth+1;
	rkid->lc = mc, rkid->rc = new->rc, rkid->lv = mv, rkid->rv = new->rv;
	LoopBody_h(rkid);

	/* NOTHREAD */
	lkid->depth = new->depth+1;
	lkid->lc = new->lc, lkid->rc = mc, lkid->lv = new->lv, lkid->rv = mv;
	LoopBody_h(lkid);
    }

}



/*
 * Mimd
 *----------------------------------------------------------------------------
 */

void Mimd(void)
{
   				 /* enqueue root task */
    double gl, gu;
    Interval_t * root;

    Gerschgorin(n, d, e, &gl, &gu);
    root = malloc(sizeof(Interval_t));
    root->depth = 0, root->lc = 0, root->rc = n, root->lv = gl, root->rv = gu;

    LoopBody_h(root);
    
}


/*
 * main
 *----------------------------------------------------------------------------
 */

void main( int argc, char ** argv )
{
    struct timeval t;

    char   c;

    gettimeofday(&t,NULL);
    srand(123);
/*    srand((unsigned int) (t.tv_sec+t.tv_usec));       /* init rnd generator */



    if ( argc < 3 ) {
	printf("usage: %s init size\n", argv[0]);
	printf("       init = r)andom, g)eometric or u)niform\n");
	printf("       size = size of matrix\n");
	exit(1);
    }
    c = *argv[1];
    n = atoi(argv[2]);

    d = malloc(n * sizeof(double));		/* allocate matrix */
    e = malloc(n * sizeof(double));

    switch (c) {				/* initialize matrix */
    case 'r': FillMatrixRnd(n, d, e); break;
    case 'g': FillMatrixGeom(n, d, e); break;
    case 'u': FillMatrixUnif(n, d, e); break;
    default:
	printf("1st argument must be 'r', 'g' or 'e'!\n");
	exit(2);
    }
    
    Mimd();
    exit(0);
}

