
/* Multiprocessor/Thread stuff */

#include <thread.h>
#include <synch.h>
#include <unistd.h>

extern long sysconf(int name);
int  cpu_num = 0;               /* Number of processors online */

int      numthreads = 0;        /* Number of active threads */
int      numallthreads = 0;     /* Number of threads created at all */
rwlock_t numthreads_rwlock;

/* Start new thread if... predicates (general ones) */

#define THREAD_NUMBER          3                        /* R */
#define NEW_THREAD_P_FIRST_N   (numallthreads < cpu_num*THREAD_NUMBER)
#define NEW_THREAD_P_KEEP_N    (numthreads < cpu_num*THREAD_NUMBER)
#define NEW_THREAD_P_ACTIVE_N  (numthreads < cpu_num*THREAD_NUMBER)
#define NEW_THREAD_P_ALWAYS    (1)
#define NEW_THREAD_P_NEVER     (0)
#define NEW_THREAD_P_NEVEREVER (0)

#define NEW_THREAD_P_NAME_FIRST_N   "FIRST_N"
#define NEW_THREAD_P_NAME_KEEP_N    "KEEP_N"
#define NEW_THREAD_P_NAME_ACTIVE_N  "ACTIVE_N"
#define NEW_THREAD_P_NAME_ALWAYS    "ALWAYS"
#define NEW_THREAD_P_NAME_NEVER     "NEVER"
#define NEW_THREAD_P_NAME_NEVEREVER "NEVEREVER (NO THREADSTUFF AT ALL)"



/* use this for neverever! */

#define THREADS_REALLY_DISABLED 0

/* Combine problem specific and general predicate */

/* Switches for on/off and corresponding definitions: */
/* if general predicate is used, activate r/w lock usage. */
/* choose right thread creation predicate (general/depth/combined) */

#define USE_DEPTH_P   0                                 /* R */
#define USE_GENERAL_P 1                                 /* R */

#define USE_FIRST_N  0					/* R */
#define USE_KEEP_N   0					/* R */
#define USE_ACTIVE_N 1					/* R */
#define USE_ALWAYS   0					/* R */
#define USE_NEVER    0					/* R */
#define USE_DEPTH    0					/* R */

#if USE_GENERAL_P && (USE_FIRST_N || USE_KEEP_N || USE_ACTIVE_N)

#define THREAD_RDLOCK rw_rdlock(&numthreads_rwlock)
#define THREAD_UNLOCK rw_unlock(&numthreads_rwlock)
#define THREAD_WRLOCK rw_wrlock(&numthreads_rwlock)

#else	/* No locking necessary for always/never/neverever */

#define THREAD_RDLOCK ;
#define THREAD_UNLOCK ;
#define THREAD_WRLOCK ;

#endif

#if USE_GENERAL_P

#if USE_FIRST_N
#define NEW_THREAD_P        NEW_THREAD_P_FIRST_N
#define NEW_THREAD_P_NAME   NEW_THREAD_P_NAME_FIRST_N
#define NEW_THREAD_P_NUMBER NEW_THREAD_P_NUMBER_FIRST_N
#elif USE_KEEP_N
#define NEW_THREAD_P        NEW_THREAD_P_KEEP_N
#define NEW_THREAD_P_NAME   NEW_THREAD_P_NAME_KEEP_N
#define NEW_THREAD_P_NUMBER NEW_THREAD_P_NUMBER_KEEP_N
#elif USE_ACTIVE_N
#define NEW_THREAD_P        NEW_THREAD_P_ACTIVE_N
#define NEW_THREAD_P_NAME   NEW_THREAD_P_NAME_ACTIVE_N
#define NEW_THREAD_P_NUMBER NEW_THREAD_P_NUMBER_ACTIVE_N
#elif USE_ALWAYS
#define NEW_THREAD_P        NEW_THREAD_P_ALWAYS
#define NEW_THREAD_P_NAME   NEW_THREAD_P_NAME_ALWAYS
#define NEW_THREAD_P_NUMBER NEW_THREAD_P_NUMBER_ALWAYS
#elif USE_NEVER
#define NEW_THREAD_P        NEW_THREAD_P_NEVER
#define NEW_THREAD_P_NAME   NEW_THREAD_P_NAME_NEVER
#define NEW_THREAD_P_NUMBER NEW_THREAD_P_NUMBER_NEVER
#else
#error "unknown general predicate - check USE_XXX!"
#endif

#endif

/* Problem specific: recursion depth until which to create threads	*/
/* (for the procedure name from profile_depth_XXX)			*/


#define DEPTH_NUMBER_compute_area        3                       /* R */
#define NEW_THREAD_P_DEPTH_compute_area  (profile_depth_compute_area < DEPTH_NUMBER_compute_area)

#if USE_GENERAL_P && USE_DEPTH_P
#define NEW_THREAD_PREDICATE_compute_area (NEW_THREAD_P_DEPTH_compute_area && NEW_THREAD_P)
#endif
#if USE_GENERAL_P && !USE_DEPTH_P
#define NEW_THREAD_PREDICATE_compute_area NEW_THREAD_P
#endif
#if !USE_GENERAL_P && USE_DEPTH_P
#define NEW_THREAD_PREDICATE_compute_area NEW_THREAD_P_DEPTH_compute_area
#endif
#if !USE_GENERAL_P && !USE_DEPTH_P
#error "at least one of USE_GENERAL_P and USE_DEPTH_P must be true!"
#endif



/*
 * Most Recent Version 29-oct-97:
 *  - no more debugging options etc
 *  - choice of 5 different pictures (07-nov-97)
 *
 *
 * Fractal  UNINSTRUMENTIERT
 *
 * Parallel Recursive Benchmark - compute Mandelbrot set using
 * "if edges have same color, fill it" heuristic and Divide&Conquer
 * (computes e.g. just 95886 pixels on a 512x512 fractal instead of 262144 in
 * the brute force approach)
 *
 *
 * Usage:
 *
 * fractal  [x_pixel [y_pixel [recurs]]]
 *
 *		x_pixel	= number of pixels to compute (x)
 *		y_pixel = dito for y
 *		recurs	= 0 for naive algorithm, otherwise recursive D&C
 *
 *
 * Stefan U. Haenssgen  27-jan-97
 *
 *
 * 27-jan-97	1st version, rough implementation
 *		Use x1,y1-x2,y2 instead of x,y,size
 *		Use independent x/y size
 *		Use recursive/naive algorithm dependent on argv
 * 28-jan-97	Debug level flexible, output ASCII data if large enough etc
 * 29-jan-97	Added instrumentation and profiling
 *		 (ProfileAddStats, ...)
 *		Added Thread creation, predicates etc
 * 31-jan-97	Added timing around computation itself, i.e. don't
 *		 measure malloc!
 *		Removed pixel counter, isn't mutex protected, may cause
 *		 write conflicts, just ignore it...
 *
 */



#include <stdio.h>

#include <sys/time.h>
#include <sys/timeb.h>



#define MAXITER 1024			/* Maximum recursion depth	*/

unsigned char *pixel;			/* Pixel values 		*/

/* Upper left coordinates and Coordinate size (x and y) */

#define COORDS 1

#if COORDS == 1
double xcorner = -2.25;                 /* basic fractal, symmetrical   */
double ycorner = -1.5;
double csize   =  3.0;
#endif
#if COORDS == 2
double xcorner = -0.058500;             /* much work in lower left corner*/
double ycorner = -0.885000;
double csize   =  0.120000;
#endif
#if COORDS == 3
double xcorner = -0.013095;             /* only lower right corner      */
double ycorner = -0.810949;
double csize   =  0.000038;
#endif
#if COORDS == 4
double xcorner = -0.538020;             /* non-symmetrical but equal work*/
double ycorner = -0.528120;
double csize   =  0.004800;
#endif
#if COORDS == 5
double xcorner = -0.764412;             /* little work in lower right   */
double ycorner = -0.101004;
double csize   =  0.000960;
#endif

int    pxsize  = 60;			/* Pixel number x, y		*/
int    pysize  = 30;
double gap     =  0.1;			/* Distance between screen pixels */
int    iter    = MAXITER;		/* Iteration limit		*/

int    use_recursion = 1;		/* Use recursive algorithm?	*/

/*long cp = 0;				/* Computed pixels counter	*/



/* Declare recursive function here */

void compute_area(int profile_depth_compute_area,int x1, int y1, int x2, int y2);


/* Output graphics as ASCII picture on stdout */

void print_pixels()
{
    int x,y;
	
    for (y=0; y<pysize; y++) {
	for (x=0; x<pxsize; x++) {
	    putchar(pixel[x+pxsize*y]%64+32);
	}	
	putchar('\n');
	fflush(stdout);
    }
}
    

/* Compute one pixel at the given pixel coordinates */

unsigned char compute_pixel(int x, int y)
{
    int count;
    double xc, yc, xt, yt, xz, yz, t;

    /*cp++;*/
    
    xc = xcorner + ((double)x)*gap;
    yc = ycorner + ((double)y)*gap;
    
    xt=yt=xz=yz=count=0;
    while ((++count<iter) && (xt+yt<4)) {
	t  = xz*yz;
	yz = t + t + yc;
	xz = xt-yt+xc;
	xt = xz*xz;
	yt = yz*yz;
    }

    return((unsigned char) count % 256);
}


/* Compute edges of area (x1,y1) to (x2, y2) */
/* size has to be >= 1 */

void compute_frame(int x1, int y1, int x2, int y2)
{
    int xx,yy;

    yy=y1;
    for (xx=x1; xx<=x2; xx++)
	pixel[xx+pxsize*yy] = compute_pixel(xx,yy);
    yy=y2;
    for (xx=x1; xx<=x2; xx++)
	pixel[xx+pxsize*yy] = compute_pixel(xx,yy);
    xx=x1;
    for (yy=y1+1; yy<=y2-1; yy++)	/* (don't compute corners twice) */
	pixel[xx+pxsize*yy] = compute_pixel(xx,yy);
    xx=x2;
    for (yy=y1+1; yy<=y2-1; yy++)
	pixel[xx+pxsize*yy] = compute_pixel(xx,yy);
    
}


/* Compute cross through area */
/* size has to be >= 1, don't re-compute edge pixels */

void compute_cross(int x1, int y1, int x2, int y2)
{
    int xx,yy,y0;

    yy=(y1+y2)/2;
    for (xx=x1+1; xx<=x2-1; xx++)
	pixel[xx+pxsize*yy] = compute_pixel(xx,yy);
    y0=yy;			/* don't compute middle pixel twice */
    xx=(x1+x2)/2;
    for (yy=y1+1; yy<=y0-1; yy++)
	pixel[xx+pxsize*yy] = compute_pixel(xx,yy);
    for (yy=y0+1; yy<=y2-1; yy++)
	pixel[xx+pxsize*yy] = compute_pixel(xx,yy);
    
}


/* True if edges of area are same color */

char edges_match(int x1, int y1, int x2, int y2)
{
    int 	  xx,yy;
    char          match = 1;
    unsigned char color = pixel[x1+pxsize*y1];	/* upper left corner 	*/

    xx=x1;
    yy=y1;
    while ((match) && (xx<=x2)) {
	match = (color == pixel[xx+pxsize*yy]);
	xx++;
    }
    xx=x1;
    yy=y2;
    while ((match) && (xx<=x2)) {
	match = (color == pixel[xx+pxsize*yy]);
	xx++;
    }
    xx=x1;
    yy=y1+1;
    while ((match) && (yy<=y2-1)) {
	match = (color == pixel[xx+pxsize*yy]);
	yy++;
    }
    xx=x2;
    yy=y1+1;
    while ((match) && (yy<=y2-1)) {
	match = (color == pixel[xx+pxsize*yy]);
	yy++;
    }

    return(match);
}


/* Compute area (x1,y1) to (x2,y2) */
/* Fill area if all edges have same color */
/* Don't touch already computed areas (i.e. edges) */


int init_threads_compute_area(thread_t *newthreads_compute_area) {
    int nt_i;
    for (nt_i = 0; nt_i < 12; nt_i++) {
        newthreads_compute_area[nt_i] = (thread_t) 0;
    }
    return 0;
}
typedef struct {
  int profile_depth_compute_area;
  int x1;
  int y1;
  int x2;
  int y2;
} ThreadArg_compute_area;

void *ThreadWrapper_compute_area(void *arg);


/*
 * thread_t Thread_compute_area(int profile_depth_compute_area, int x1,int y1,int x2,int y2)
 *
 * Wrapper call to execute specific function as a new thread
 * with the given parameters.
 * Returns ID of the thread created, for later joining and
 * retrieving results.
 */
thread_t  Thread_compute_area(int profile_depth_compute_area, int x1,int y1,int x2,int y2)
{
    thread_t                thread_id = 0;
    ThreadArg_compute_area *arg;
    
    arg = (ThreadArg_compute_area*)malloc(sizeof(ThreadArg_compute_area));
    
    arg-> profile_depth_compute_area = profile_depth_compute_area;
    arg-> x1 = x1;
    arg-> y1 = y1;
    arg-> x2 = x2;
    arg-> y2 = y2;

	   
    if ( thr_create(NULL,0,                    /* Stack stuff */
                    ThreadWrapper_compute_area,     /* Fn to call */
                    (void*)arg,                /* Argument to Fn */
                    0,                         /* Flags */
                    &thread_id) ) {            /* New thread's ID */
        perror("Error creating thread");
        printf("%d threads created so far. Aborting\n", numallthreads);
        exit(123);
    }
    return(thread_id);
}
/*
 * void Join_compute_area(thread_id)
 *
 * Waits for given thread to finish, unpacks its result from the
 * thread's return value and returns it.
 */
void Join_compute_area(thread_t thread_id)
{
    void *result;

    thr_join(thread_id, NULL, &result);

#if USE_KEEP_N
    THREAD_WRLOCK;
    numthreads--;
    THREAD_UNLOCK;
#endif

    return;
}


/*
 * ThreadWrapper_compute_area(arg)
 *
 * This function is used at the new thread's start.
 * It unpacks its arguments (ThreadArg struct *) and then calls
 * the specific function with the correct arguments
 */
void *ThreadWrapper_compute_area(void *arg)
{

    
    
    /* Call function, get result */

    compute_area(((ThreadArg_compute_area*) arg)-> profile_depth_compute_area,
        ((ThreadArg_compute_area*) arg) -> x1,
        ((ThreadArg_compute_area*) arg) -> y1,
        ((ThreadArg_compute_area*) arg) -> x2,
        ((ThreadArg_compute_area*) arg) -> y2    
);

    
    free(arg);

#if !USE_KEEP_N
    THREAD_WRLOCK;
    numthreads--;
    THREAD_UNLOCK;
#endif

    thr_exit((void*)NULL);
}


void compute_area(int profile_depth_compute_area,int x1, int y1, int x2, int y2)
{
int branch_count = 0;

    thread_t newthreads_compute_area[12];
    int      nt_i = init_threads_compute_area(&(newthreads_compute_area[0]));

    unsigned char p;
    int           xx,yy;
    int           dx = x2-x1;
    int           dy = y2-y1;

    
    if ((dx < 2) || (dy < 2)) {
    ;
#if THREADS_REALLY_DISABLED
#else
    /* If thread was created, join it */
    for (nt_i = 0; nt_i < branch_count; nt_i++) {
    
    if (newthreads_compute_area[nt_i]) {
            Join_compute_area(newthreads_compute_area[nt_i]);
            newthreads_compute_area[nt_i] = (thread_t) 0; /* Mark thread as joined */
        }
    
}
#endif
    

	return;
    }
    if (dx == 2) {			/* Handle single-line cases	*/
	xx=(x1+x2)/2;
	for (yy = y1+1; yy <= y2-1; yy++)
	    pixel[xx+pxsize*yy] = compute_pixel(xx,yy);
	    ;
#if THREADS_REALLY_DISABLED
#else
	    /* If thread was created, join it */
	    for (nt_i = 0; nt_i < branch_count; nt_i++) {
	    
    if (newthreads_compute_area[nt_i]) {
	            Join_compute_area(newthreads_compute_area[nt_i]);
	            newthreads_compute_area[nt_i] = (thread_t) 0; /* Mark thread as joined */
	        }
	    
}
#endif
	    

	return;
    }
    if (dy == 2) {
	yy=(y1+y2)/2;
	for (xx = x1+1; xx <= x2-1; xx++)
	    pixel[xx+pxsize*yy] = compute_pixel(xx,yy);
	    ;
#if THREADS_REALLY_DISABLED
#else
	    /* If thread was created, join it */
	    for (nt_i = 0; nt_i < branch_count; nt_i++) {
	    
    if (newthreads_compute_area[nt_i]) {
	            Join_compute_area(newthreads_compute_area[nt_i]);
	            newthreads_compute_area[nt_i] = (thread_t) 0; /* Mark thread as joined */
	        }
	    
}
#endif
	    

	return;
    }
    if (edges_match(x1,y1,x2,y2)) {	/* If edges match, fill area - done! */
	
	p = pixel[x1+pxsize*y1];
	
	for (yy=y1+1; yy<=y2-1; yy++) {
	    for (xx=x1+1; xx<=x2-1; xx++) {
		pixel[xx+pxsize*yy] = p;
	    }
	}
	;
#if THREADS_REALLY_DISABLED
#else
	/* If thread was created, join it */
	for (nt_i = 0; nt_i < branch_count; nt_i++) {
	
    if (newthreads_compute_area[nt_i]) {
	        Join_compute_area(newthreads_compute_area[nt_i]);
	        newthreads_compute_area[nt_i] = (thread_t) 0; /* Mark thread as joined */
	    }
	
}
#endif
	

	return;
    }

    /* Subdivide area into quadrants, compute them recursively 		*/
    
    compute_cross(x1,y1,x2,y2);
#if THREADS_REALLY_DISABLED
    ;
    
    {
    branch_count++;
    compute_area(profile_depth_compute_area+1, x1, y1, (x1+x2)/2, (y1+y2)/2);}
    ;
#else
        /* Create new thread if predicate tells us to */
        /* (use rwlock because predicate may read number-of-threads etc) */
    
    THREAD_RDLOCK;
        if (NEW_THREAD_PREDICATE_compute_area) {     /* Condition GO */
            THREAD_UNLOCK;
            THREAD_WRLOCK;                      /* Now serious */
            if (NEW_THREAD_PREDICATE_compute_area) { /* Condition still OK? */
                numallthreads++;
                numthreads++;                   /* OK, new thread! */
                THREAD_UNLOCK;
                branch_count++;
                newthreads_compute_area[branch_count-1] = Thread_compute_area(profile_depth_compute_area+1, x1, y1, (x1+x2)/2, (y1+y2)/2);
            } else {                            /* Someone else created one */
                THREAD_UNLOCK;                  /* so we can't any more */
                {
                branch_count++;
                compute_area(profile_depth_compute_area+1, x1, y1, (x1+x2)/2, (y1+y2)/2);}
                ;
            }
        } else {                                /* Condition not OK, */
            THREAD_UNLOCK;                      /* perform serial call */
            {
            branch_count++;
            compute_area(profile_depth_compute_area+1, x1, y1, (x1+x2)/2, (y1+y2)/2);}
            ;
        }
#endif
    
;
#if THREADS_REALLY_DISABLED
    ;
    {
    branch_count++;
    compute_area(profile_depth_compute_area+1, (x1+x2)/2, y1, x2, (y1+y2)/2);}
    ;
#else
        /* Create new thread if predicate tells us to */
        /* (use rwlock because predicate may read number-of-threads etc) */
    
    THREAD_RDLOCK;
        if (NEW_THREAD_PREDICATE_compute_area) {     /* Condition GO */
            THREAD_UNLOCK;
            THREAD_WRLOCK;                      /* Now serious */
            if (NEW_THREAD_PREDICATE_compute_area) { /* Condition still OK? */
                numallthreads++;
                numthreads++;                   /* OK, new thread! */
                THREAD_UNLOCK;
                branch_count++;
                newthreads_compute_area[branch_count-1] = Thread_compute_area(profile_depth_compute_area+1, (x1+x2)/2, y1, x2, (y1+y2)/2);
            } else {                            /* Someone else created one */
                THREAD_UNLOCK;                  /* so we can't any more */
                {
                branch_count++;
                compute_area(profile_depth_compute_area+1, (x1+x2)/2, y1, x2, (y1+y2)/2);}
                ;
            }
        } else {                                /* Condition not OK, */
            THREAD_UNLOCK;                      /* perform serial call */
            {
            branch_count++;
            compute_area(profile_depth_compute_area+1, (x1+x2)/2, y1, x2, (y1+y2)/2);}
            ;
        }
#endif
    
;
#if THREADS_REALLY_DISABLED
    ;
    {
    branch_count++;
    compute_area(profile_depth_compute_area+1, x1, (y1+y2)/2, (x1+x2)/2, y2);}
    ;
#else
        /* Create new thread if predicate tells us to */
        /* (use rwlock because predicate may read number-of-threads etc) */
    
    THREAD_RDLOCK;
        if (NEW_THREAD_PREDICATE_compute_area) {     /* Condition GO */
            THREAD_UNLOCK;
            THREAD_WRLOCK;                      /* Now serious */
            if (NEW_THREAD_PREDICATE_compute_area) { /* Condition still OK? */
                numallthreads++;
                numthreads++;                   /* OK, new thread! */
                THREAD_UNLOCK;
                branch_count++;
                newthreads_compute_area[branch_count-1] = Thread_compute_area(profile_depth_compute_area+1, x1, (y1+y2)/2, (x1+x2)/2, y2);
            } else {                            /* Someone else created one */
                THREAD_UNLOCK;                  /* so we can't any more */
                {
                branch_count++;
                compute_area(profile_depth_compute_area+1, x1, (y1+y2)/2, (x1+x2)/2, y2);}
                ;
            }
        } else {                                /* Condition not OK, */
            THREAD_UNLOCK;                      /* perform serial call */
            {
            branch_count++;
            compute_area(profile_depth_compute_area+1, x1, (y1+y2)/2, (x1+x2)/2, y2);}
            ;
        }
#endif
    
;
    {
    branch_count++;
    compute_area(profile_depth_compute_area+1, (x1+x2)/2, (y1+y2)/2, x2, y2);}
    ; /* NOTHREAD */
    

;
#if THREADS_REALLY_DISABLED
#else
/* If thread was created, join it */
for (nt_i = 0; nt_i < branch_count; nt_i++) {

    if (newthreads_compute_area[nt_i]) {
        Join_compute_area(newthreads_compute_area[nt_i]);
        newthreads_compute_area[nt_i] = (thread_t) 0; /* Mark thread as joined */
    }

}
#endif

}


void PrintThreadNumberReachedAndInfo() {
   printf("BEGIN REAPAR THREADINFO\n");
   printf("total number of threads created = %d\n",
          numallthreads);
   printf("END REAPAR THREADINFO\n");
}

int do_thread_init(){


    printf("BEGIN REAPAR PARINFO\n");

    /* Thread stuff: get number of processors, initialize
     * number-of-threads counters and mutex protecting them
     */
    atexit(PrintThreadNumberReachedAndInfo);
    cpu_num = (int)sysconf(_SC_NPROCESSORS_ONLN);
        { char *ev;
          if ((ev = (char*)getenv("T_CPU_NUM")) != NULL) {
            if ((atoi(ev) > cpu_num) || (atoi(ev)<1)) {
              printf("Error, %d processors requested but just %d available\n!",
                     atoi(ev), cpu_num);
              exit(815);
            } else {
              cpu_num = atoi(ev);
            }
          }
        }
    printf("%d processor%s online\n", cpu_num, cpu_num>1?"s":"");
#if THREADS_REALLY_DISABLED
    printf("threads disabled\n");
#else
    thr_setconcurrency((int)cpu_num);
    rwlock_init(&numthreads_rwlock, USYNC_THREAD, NULL);
#endif
    numthreads = 0;
    numallthreads = 0;
    
#if USE_GENERAL_P
    printf("strategy general   = %s (N=%d)\n", NEW_THREAD_P_NAME,
           THREAD_NUMBER);
#endif
#if USE_DEPTH_P
    printf("strategy max depth (compute_area) = %d\n", DEPTH_NUMBER_compute_area);
#endif

printf("\n");
  printf("Parallel procedure (%2d) %-20s at line %4d\n",
         1, "compute_area", 226);
printf("END REAPAR PARINFO\n");}


main(int argc, char **argv)
{
  int dummy_thread_init = do_thread_init();

    struct timeval starttime;
    struct timeval endtime;
    int x,y;
    int i;

    
    i=1;
    if (argc>i) {
	pxsize = atoi(argv[i]);
    }
    i++;
    if (argc>i) {
	pysize = atoi(argv[i]);
    }
    i++;
    if (argc>i) {
	use_recursion = atoi(argv[i]);
    }
    i++;

    printf("Fractal: %d x %d\n", pxsize, pysize);

    gap = csize/pxsize;			/* Distance between screen pixels */

    /* Allocate & initialize pixel buffer */
    
    pixel = (unsigned char*) malloc(sizeof(unsigned char) *
				    pxsize * pysize + 1);
    for (x=0; x<pxsize*pysize; x++)
	pixel[x]=(unsigned char)0;


    gettimeofday(&starttime,NULL);
    
    if (use_recursion) {

	printf("Using recursive algorithm\n");
	
	compute_frame(0,0,pxsize-1,pysize-1);

	compute_area(0, 0,0,pxsize-1,pysize-1);

    } else {

	printf("Using naive (non-recursive) algorithm\n");
    
	for (y=0; y<pysize; y++) {
	    for (x=0; x<pxsize; x++) {
		pixel[x+pxsize*y] = compute_pixel(x,y);
	    }
	}
    
    }

    gettimeofday(&endtime,NULL);

    printf("Elapsed time %f\n",
	   (endtime.tv_sec - starttime.tv_sec) +
	   (endtime.tv_usec - starttime.tv_usec)/1000000.0 );

    /*printf("%d pixels computed\n",cp);*/
    
    exit(0);

}
