/*
 * UNINSTRUMENTIERT ABER SCHON MIT UMSTRUKTURIERUNG
 * (OK-Check erst in Rekursion, nicht vorher)
 *
 * queens - A simple nqueens program to test (recursive parallel benchmark)
 *	    Generate all (!) solutions to queens problem on NxN board
 *
 * Adapted from Cilk sample program.
 *
 * usage:	queens n [output]
 *		output number of solutions for n*n board
 *		if output is !=0, output solution themselves, too
 *
 * Stefan U. Haenssgen  09-feb-97
 *
 * 09-feb-97	Adapted from Cilk sample program:
 *		Removed "cilk", "catch" etc
 *		Enhanced comments
 *		Collect ALL results, not just one
 *		Use struct+union to collect results
 *		Added print_result to view all results
 *		Added profiling instrumentation
 *		Added Thread Parallelization
 * 11-feb-97	Optionally just output number of solutions found
 * 01-mar-97	Read number of CPUs to use from environment T_CPU_NUM
 * 19-mar-97	Removed instrumentation and thread stuff
 * 26-dec-97	Moved return value to reference argument
 *		Added noparallel annotation for print tree and needresults
 *		 for nqueens
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include <alloca.h>

int output = 0;

#undef DEBUG

#define LEAF 1
#define NODE 2



/* Struct for holding the result, either leaf (one solution to problem) */
/* or node with sub-solutions */
typedef struct result {
    char type;				/* LEAF or NODE */
    union {
	char           *one_result;	/* Leaf: Result for this leaf */
	struct result **results;	/* Node: all N sub-results */
    } r;
} Result;

    
/*
 * <a> contains array of <n> queen positions.  Returns 1
 * if none of the queens conflict, and returns 0 otherwise.
 */
int ok(int n, char *a)
{
    int i, j;
    char p, q;

    for (i = 0; i < n; i++) {
	p = a[i];
	
	for (j = i + 1; j < n; j++) {
	    q = a[j];
	    if (q == p || q == p - (j - i) || q == p + (j - i))
		return 0;
	}
    }
    return 1;
}


/* Output result tree */

int print_result_tree(int n, Result *a, int d) /* NOPARALLEL */
{
    int i,j;
    int count=0;

    if (a==NULL) {
#ifdef DEBUG
	for (j=0; j<d; j++) printf("   ");
	printf("<>\n");
#endif
	return 0;
    }
    
    if (a->type == LEAF) {
#ifdef DEBUG
	for (j=0; j<d; j++) printf("   ");
#endif
	if (!ok(n, a->r.one_result))
	    printf("RESULT IS BAD!!!\n");
	if (output) {
	    for (j = 0; j < n; j++)
		printf("%2d ", a->r.one_result[j]);
	    printf("\n");
	}
	count++;
    } else {
	for (i=0; i<n; i++) {
	    count = count + print_result_tree(n, a->r.results[i], d+1);
	}
    }
    return count;
}

int print_result(int n, Result *a)
{
    return(print_result_tree(n, a, 0));
}



/*
 * <a> is an array of <j> numbers.  The entries of <a> contain
 * queen positions already set.  If there is any extension of <a>
 * to a complete <n> queen setting, returns all of these queen
 * settings in an array[n] (allocated from the heap).  Otherwise,
 * returns NULL.
 * Does not side-effect <a>.
 */

void nqueens(Result **outres, int n, int j, char *a)
{
     char   *b;
     Result *done = NULL;	/* Heap-allocated result or NULL */
     Result **res = NULL;	/* Temporary results of recursive calls */
     char   any;
     int    i;

     /* Managed to place all n queens? */
     /* Put good solution in heap, return pointer to it. */
     /* (set type to NODE) */

#ifdef DEBUG
     printf("nqueens(%d, %d, '", n, j);
     for (i=0; i<j; i++)
	 printf("%d",a[i]);
     printf("') \n");
#endif

     if (!ok(j, a)) {
	 *outres = NULL;
	 return;
     }

     if (n == j) {
	 b = (char*) malloc(n * sizeof(char));	/* Our solution */
	 memcpy(b, a, n * sizeof(char));
	 done = (Result*) malloc(sizeof(Result));
	 done->type = LEAF;
	 done->r.one_result = b;
	 *outres = done;
	 return;
     }

     /* Try each possible position for queen <j>: */
     /* Allocate a temporary array on stack and copy <a> into it */
     
     res = (Result**) alloca(n * sizeof(Result*));
     for (i = 0; i < n; i++) {
	 b = alloca((j + 1) * sizeof(char));
	 memcpy(b, a, j * sizeof(char));
	 b[j] = i;
	     
	 nqueens(&(res[i]), n, j + 1, b);
     }

     /* NEEDRESULTS */
				 
     /* OK, now check if any solutions came up. Yes -> remember them */
     /* No -> return NULL */
     
     any=0;
     for (i = 0; i < n; i++) {
	 if (res[i])
	     any=1;
     }
     if (any) {
	 done = (Result*) malloc(sizeof(Result));
	 done->type = NODE;
	 done->r.results = (Result**) malloc(n * sizeof(Result*));
	 for (i=0; i<n; i++) {			/* Copy solutions to heap */
	     if (res[i]) {
		 done->r.results[i] = (Result*)malloc(sizeof(Result));
		 done->r.results[i]->type = res[i]->type;
		 done->r.results[i]->r    = res[i]->r;
	     } else {
		 done->r.results[i] = (Result*)NULL;
	     }
	 }
     }
     *outres = done;
     return;

}


void main(int argc, char *argv[])
{
     int n;
     char *a;
     Result *res;
     int i;
     int count;

     if (argc < 2) {
	  printf("%s: number of queens required\n", argv[0]);
	  exit(1);
     }
     if (argc > 3)
	  printf("%s: extra arguments being ignored\n", argv[0]);

     n = atoi(argv[1]);
     a = alloca(n * sizeof(char));

     if (argc>2) {
	 output = atoi(argv[2]);
     }

     printf("running queens %d\n", n);
     nqueens(&res, n, 0, a);

     if (res != NULL) {
	 count = print_result(n,res);
	 printf("\n%d solutions found\n", count);
     } else
	 printf("no possible result!\n");
     
     exit(0);
}
