/*
 * profile.h
 *
 * basic definitions for profiling
 *
 * Stefan U. Haenssgen  27-nov-96
 *
 *
 * 27-nov-96	1st version
 *		Collect statistics in
 *		  array of lines, each with
 *		    array of depths reached, each with
 *		      array of degree-counts encountered
 *		i.e. line -> depth -> degree  counts how often this
 *		 line at this recursion depth had this branching degree
 *		 (degree 0 means it was a leaf)
 *		Renamed variables
 * 28-nov-96	ProfileNewDefaultStats() for initialization,
 *		 initial sizes as parameters for ProfileNewStats()
 *		New field "info" for line stats
 *
 */

#ifndef this_is_profile_h
#define this_is_profile_h

#include <stdio.h>

#define INITIALLINE	1000	/* Initial max number of lines in program */
#define INITIALDEPTH	20	/* Initial size of depths-array for a line */
#define INITIALDEGREE	5	/* Dito for degree array */

typedef struct {
    int   	      maxdegree;	/* Current size of count-array */
    long             *degrees;		/* Counts for each degree */
} profile_depth_t;

typedef struct {
    char	     *info;		/* Information (names, etc) */
    int               maxdepth;		/* Current size of depth-array */
    profile_depth_t  *depths;		/* Degrees for each depth */
} profile_line_t;

typedef struct {
    int  	      initialdepth;	/* Default initial sizes */
    int 	      initialdegree;
    int 	      maxlines;		/* Current size of line-array */
    profile_line_t   *lines;		/* Depths for each line */
} profile_stat_t;

#define ProfileAddStat(pstats, pline, pdepth, pdegree, pcount) \
(pstats->lines[pline].depths[pdepth].degrees[pdegree])+= (long)pcount


#ifndef profile_c
extern profile_stat_t *profile_stat;
extern profile_stat_t *ProfileNewStats(int lines, int depths, int degrees);
extern profile_stat_t *ProfileNewDefaultStats(void);
extern void ProfileInitLine(profile_stat_t *stats, int line, char *info);
extern void ProfilePrint(profile_stat_t *stats);
extern void ProfileFree(profile_stat_t *stats);
#endif

#endif
